module java.util.HashMap;

import java.lang.all;
import java.util.Map;
import java.util.Collection;
import java.util.Set;
import java.util.HashSet;
import java.util.ArrayList;

static import tango.util.container.HashMap;

private struct ObjRef {
    Object obj;
    static ObjRef opCall( Object obj ){
        ObjRef res;
        res.obj = obj;
        return res;
    }
    public hash_t toHash(){
        return obj is null ? 0 : obj.toHash();
    }
    public int opEquals( ObjRef other ){
        return obj is null ? other.obj is null : obj.opEquals( other.obj );
    }
    public int opEquals( Object other ){
        return obj is null ? other is null : obj.opEquals( other );
    }
}

class HashMap : Map {
    // The HashMap  class is roughly equivalent to Hashtable, except that it is unsynchronized and permits nulls.
    alias tango.util.container.HashMap.HashMap!(ObjRef,ObjRef) MapType;
    private MapType map;

    public this(){
        map = new MapType();
    }
    public this(int initialCapacity){
        this();
    }
    public this(int initialCapacity, float loadFactor){
        map = new MapType(loadFactor);
    }
    public this(Map m){
        this();
        putAll(m);
    }
    public void clear(){
        map.clear();
    }
    public bool containsKey(Object key){
        ObjRef v;
        ObjRef keyr = ObjRef(key);
        return map.get(keyr, v );
    }
    public bool containsKey(String key){
        return containsKey(stringcast(key));
    }
    public bool containsValue(Object value){
        ObjRef valuer = ObjRef(value);
        return map.contains(valuer);
    }
    public Set  entrySet(){
        HashSet res = new HashSet();
        foreach( k, v; map ){
            res.add( new MapEntry(this,k.obj));
        }
        return res;
    }
    public override int opEquals(Object o){
        if( auto other = cast(HashMap) o ){
            if( other.size() !is size() ){
                return false;
            }
            foreach( k, v; map ){
                auto vo = other.get(k.obj);
                if( v != vo ){
                    return false;
                }
            }
            return true;
        }
        return false;
    }
    public Object get(Object key){
        ObjRef keyr = ObjRef(key);
        if( auto v = keyr in map ){
            return (*v).obj;
        }
        return null;
    }
    public Object get(String key){
        return get(stringcast(key));
    }
    public override hash_t toHash(){
        return super.toHash();
    }
    public bool isEmpty(){
        return map.isEmpty();
    }
    public Set    keySet(){
        HashSet res = new HashSet();
        foreach( k, v; map ){
            res.add(k.obj);
        }
        return res;
    }
    public Object put(Object key, Object value){
        ObjRef valuer = ObjRef(value);
        ObjRef keyr = ObjRef(key);
        Object res = null;
        if( auto vold = keyr in map ){
            res = (*vold).obj;
        }
        map[ keyr ] = valuer;
        return res;
    }
    public Object put(String key, Object value){
        return put( stringcast(key), value );
    }
    public Object put(Object key, String value){
        return put( key, stringcast(value) );
    }
    public Object put(String key, String value){
        return put( stringcast(key), stringcast(value) );
    }
    public void   putAll(Map t){
        foreach( k, v; t ){
            map[ObjRef(k)] = ObjRef(v);
        }
    }
    public Object remove(Object key){
        ObjRef keyr = ObjRef(key);
        if( auto v = keyr in map ){
            Object res = (*v).obj;
            map.remove(keyr);
            return res;
        }
        map.remove(keyr);
        return null;
    }
    public Object remove(String key){
        return remove(stringcast(key));
    }
    public int    size(){
        return map.size();
    }
    public Collection values(){
        ArrayList res = new ArrayList( size() );
        foreach( k, v; map ){
            res.add( v.obj );
        }
        return res;
    }

    public int opApply (int delegate(ref Object value) dg){
        int ldg( ref ObjRef or ){
            return dg( or.obj );
        }
        return map.opApply( &ldg );
    }
    public int opApply (int delegate(ref Object key, ref Object value) dg){
        int ldg( ref ObjRef key, ref ObjRef value ){
            return dg( key.obj, value.obj );
        }
        return map.opApply( &ldg );
    }

}

